class Distributor
  module Actions

    class Subscribe < Action
      FORBIDDEN_TEMPLATE = Actions.template("subscribe.forbidden")
      SUCCESS_TEMPLATE = Actions.template("subscribe.success")
      WAIT_USER_SUBSCRIBE_TEMPLATE = Actions.template("subscribe.wait_user")
      WAIT_MODO_SUBSCRIBE_TEMPLATE = Actions.template("subscribe.wait_modo")

      def handle(list:, mail:, attributes:)
        register =
          begin
            Email.register!(mailinglist: list, name: mail.from_name, email: mail.from).save
          rescue StandardError => e
            $logger.error e.message
            nil
          end
        if register
          if !register.reader?
            handle_wait_validation(list:, mail:, register:)
          else
            handle_subscribed(list:, mail:, register:)
          end
        else
          handle_403(list:, mail:)
        end
      end

      def handle_wait_validation(list:, mail:, register:)
        $logger.debug "Subscribe#handle_wait_validation on #{list.email} for #{mail.from}"
        body = WAIT_USER_SUBSCRIBE_TEMPLATE.result binding
        @distributor.distribute(mail.to_response(list:, mail:, body:))

        modo = list.enabled_modos.first # TODO: send to all
        body = WAIT_MODO_SUBSCRIBE_TEMPLATE.result binding
        @distributor.distribute(
          Protocols::Mail.build(
            subject: "ML #{list.name} requires validation for #{mail.from}", list:, to: modo.email, body:,
          ),
        )
      end

      def handle_subscribed(list:, mail:, register:)
        $logger.debug "Subscribe#handle_subscribed on #{list.email} for #{mail.from}"
        $logger.debug register.inspect
        body = SUCCESS_TEMPLATE.result binding
        @distributor.distribute(mail.to_response(list:, mail:, body:))
      end

      def handle_403(list:, mail:)
        $logger.debug "Subscribe#handle_403 on #{list.email} for #{mail.from}"
        body = FORBIDDEN_TEMPLATE.result binding
        @distributor.distribute(mail.to_response(list:, mail:, body:))
      end
    end

    class Unsubscribe < Action
      SUCCESS_TEMPLATE = Actions.template("unsubscribe.success")

      def handle(list:, mail:, attributes:)
        Email.unregister!(mailinglist: list, email: mail.from)
        body = SUCCESS_TEMPLATE.result binding
        @distributor.distribute(mail.to_response(list:, mail:, body:))
      end
    end

    class Help < Action
      HELP_TEMPLATE = Actions.template("help")

      def handle(list:, mail:, attributes:)
        body = HELP_TEMPLATE.result binding
        @distributor.distribute(mail.to_response(list:, mail:, body:))
      end
    end

    # This distribute the mail among the readers
    class Distribute < Action
      WAIT_MODO_DISTRIBUTE_TEMPLATE = Actions.template("distribute.wait_modo")

      def handle(list:, mail:, attributes:)
        user = Email.first(mailinglist: list, email: mail.from)
        if list.registration?("autoregister")
          if user.nil?
            $logger.info "registering <#{mail.from}> to <#{list.email}>"
            Email.register!(mailinglist: list, name: mail.from_name || mail.from.split("@").first,
                            email: mail.from).save
          end
          # ok let the mail pass the security check
        elsif !user&.writer?
          $logger.warn "invalid email writer for #{mail.from} on #{mail.to}"
          return nil
        end

        if list.moderation?("restrictedwrite") && !user.modo? && !user.op?
          handle_moderated_list(list:, mail:, attributes:)
        else
          list.enabled_readers.each do |reader|
            to_distrib = mail.to_redistribute(list:, dest: reader)
            @distributor.distribute(to_distrib)
          end
        end
      end

      # @param list [Mailinglist]
      # @parma mail [Protocols::Mail]
      def handle_moderated_list(list:, mail:, attributes:)
        message_id = mail.message_id
        @distributor.imap_client.moderate mail
        uid = @distributor.imap_client.search_message_id(message_id, inbox: Protocols::Imap::MODERATION_INBOX)
        modo = list.enabled_modos.first
        # TODO: this is the most lazy code I wrote this week I think
        # if multipart, get the boundary to add a new part in the top
        boundary = mail.header("Content-Type").to_s.match(/boundary="([^"]+)"/)[1] rescue nil
        body = WAIT_MODO_DISTRIBUTE_TEMPLATE.result binding
        modo_mail = Protocols::Mail.build(
          subject: "ML #{list.name} requires validation for [#{mail.subject}]",
          list:, to: modo.email, body:,
          replace_headers: mail.kept_headers,
        )
        @distributor.distribute(modo_mail)
      end
    end

  end
end
