require "uuid"

class Mailinglist < Sequel::Model($db)
  SUFFIX_SEPARATOR = ENV["MAILINGLIST_SUFFIX_SEPARATOR"] || "."
  BASE_USER = ENV["MAILINGLIST_BASE_USER"] || "mailinglist"
  HOST = ENV["MAILINGLIST_HOST"]

  STRATEGIES = {
    registration: %w[autoregister free validated closed],
    moderation: %w[freewrite restrictedwrite],
  }.freeze

  one_to_many :emails

  def before_create
    self.created_at ||= Time.now
    super
  end

  def before_save
    self.updated_at ||= Time.now
    super
  end

  def self.build(name:, aliasing: nil, suffix: nil, strategy: "closed|freewrite")
    email = BASE_USER.dup
    email << SUFFIX_SEPARATOR << suffix if suffix
    aliasing = UUID.generate if !suffix && !aliasing
    email << aliasing if aliasing
    email << "@#{HOST}"

    new(
      name: name,
      email: email,
      count_handled: 0,
      count_distributed: 0,
      last_email: nil,
      strategy: strategy,
    )
  end

  def self.search_mail(mail)
    find(email: mail.to)
  end

  def signature
    "\r\n\r\n---\r\nYou can unsubscribe to the mailinglist #{name} via: <mailto:#{unsubscribe_email}"
  end

  def registration
    strategy.split("|").filter { STRATEGIES[:registration].include?(_1) }
  end

  def registration?(find)
    registration.include?(find)
  end

  def moderation
    strategy.split("|").filter { STRATEGIES[:moderation].include?(_1) }
  end

  def moderation?(find)
    moderation.include?(find)
  end

  def enabled_readers
    emails.filter { _1.permissions & Email::Permissions::READ != 0 }
  end

  def enabled_writers
    emails.filter { _1.permissions & Email::Permissions::WRITE != 0 }
  end

  def enabled_ops
    emails.filter { _1.permissions & Email::Permissions::OP != 0 }
  end

  def enabled_modos
    emails.filter { _1.permissions & Email::Permissions::MODO != 0 }
  end

  def actions_emails
    ACTIONS_EMAILS.to_h { [_1, send("#{_1}_email")] }
  end

  ACTIONS_EMAILS = %i[
    help subscribe unsubscribe owner
    set_permissions list_users
  ]
  def help_email
    "#{email}?subject=help"
  end

  def subscribe_email
    "#{email}?subject=subscribe"
  end

  def unsubscribe_email
    "#{email}?subject=unsubscribe"
  end

  def post_email
    email
  end

  def owner_email
    enabled_ops.first.email
  end

  def validate_distribute_email(uid)
    "#{email}?subject=validate,uid=#{uid}"
  end

  def refuse_distribute_email(uid)
    "#{email}?subject=refuse,uid=#{uid}"
  end

  def set_permissions_email(*permissions_symbols, user_email: nil, permissions: nil)
    permissions = Email::Permissions.from_symbols(*permissions_symbols) if permissions.nil?
    user_part = user_email ? ",user-email=#{user_email}" : ""
    permissions_part = ",permissions=#{permissions}"
    "#{email}?subject=set-permissions#{user_part}#{permissions_part}"
  end

  def list_users_email
    "#{email}?subject=list-users"
  end

end
