require "uuid"

# Protocls::Mail provides an interface to read the content,
# reply and redistribute the mail.
# It handles multiple headers (having several times the same header key)
# but you must know that gmail seem to do not allow it.
class Protocols::Mail
  attr_accessor :from_name, :from, :to_name, :to, :subject, :body, :uid, :message_id

  DATE_FORMAT = "%a, %d %b %Y %H:%M:%S %z".freeze
  HEADERS_KEEP = %w[
    Content-Type Content-Transfer-Encoding MIME-Version
  ].freeze
  HEADERS_KEEP_WITH_ORIGINAL = HEADERS_KEEP + %w[Message-ID User-Agent From To Subject Subject Date In-Reply-To References]
  USER_AGENT = "Ruby/#{RUBY_VERSION} (#{RUBY_PLATFORM}) Mailinglist.rb/1.0".freeze

  # @params imap_mail is any element outputed in the returned array of
  #   Net::IMAP#fetch, and must contains some attributes:
  #   - Protocols::ENVELOPE
  #   - Protocols::HEADERS
  #   - Protocols::BODYTEXT
  #   - Protocols::UID
  def initialize(imap_mail: nil)
    if imap_mail
      @imap_mail  = imap_mail
      envelope    = imap_mail.attr[Protocols::ENVELOPE]
      body_head   = imap_mail.attr[Protocols::HEADERS]
      body_text   = imap_mail.attr[Protocols::BODYTEXT]
      @from_name  = envelope.from.first.name
      @to_name    = envelope.to.first.name
      @from       = "#{envelope.from.first.mailbox}@#{envelope.from.first.host}"
      @to         = "#{envelope.to.first.mailbox}@#{envelope.to.first.host}"
      @subject    = envelope.subject
      @body       = body_text
      @headers    = Protocols::Mail.parse_rfc822_headers(body_head)
      @headers.filter! { HEADERS_KEEP_WITH_ORIGINAL.include?(_1[0]) }
      @seq        = header("X-Sequence").to_i
      @message_id = header("Message-id")
      @uid        = imap_mail.attr[Protocols::UID]
    end
  end
  attr_reader :imap_mail if $debug

  # Copy an existing mail.
  # It will NOT work if the email do not contain a imap_mail.
  def clone
    Protocols::Mail.new(imap_mail: @imap_mail)
  end

  # Interface with Protocols::Mail
  def seen!(imap_client:, inbox:)
    if @imap_mail
      imap_client.seen!(@imap_mail, inbox:)
    else
      $logger.debug("Cannot mark #{uid} as seen because no @imap_mail")
      false
    end
  end

  # Transform an received email into a mailing list email.
  # Tested with evolution and Kmail.
  # Thunderbird bug at the moment which make it weird but it works too.
  #
  # @param list [Mailinglist]
  # @param dest [Email]
  def to_redistribute(list:, dest:)
    new = clone
    new.replace_headers!(
      ["User-Agent",       USER_AGENT],
      # require people to do not respond privatly
      ["Reply-to",         "#{list.name} <#{list.email}>"],
      ["Errors-To",        list.email],
      ["To",               dest.email],
      ["From",             "\"#{@from_name}\" (via #{list.name}) <#{@from}>"],
      ["Sender",           @from],
      ["Date",             Time.now.strftime(DATE_FORMAT)],
      ["List-Id",          "<#{list.email}>"],
      ["List-Post",        "<mailto:#{list.email}>"],
      ["List-Help",        "<mailto:#{list.help_email}>"],
      ["List-Subscribe",   "<mailto:#{list.subscribe_email}>"],
      ["List-Unsubscribe", "<mailto:#{list.unsubscribe_email}>"],
      ["List-Post",        "<mailto:#{list.post_email}>"],
      ["List-Owner",       "<mailto:#{list.owner_email}>"],
      ["In-Reply-To",      @message_id],
      %w[Precedence list],
      # ["Precedence",       "bulk"],
      ["Message-Id",       "<#{UUID.generate}@#{ENV['SENDER_HOST']}>"],
      ["X-Sequence",         (@seq + 1).to_s],
      ["X-Loop",           list.email],
      ["Return-Path",      "<#{list.email}>"],
    )

    new.to   = dest.email
    new.from = list.email
    new.body = "#{body}#{list.signature}"

    new
  end

  # Create a email from scratch.
  #
  # @param list [Mailinglist]
  # @param to   [String]        email of the target
  def self.build(subject:, list:, to:, body:, replace_headers: [])
    new = new()
    new.replace_headers!(
      ["User-Agent",       USER_AGENT],
      ["Reply-to",         "#{list.name} <#{list.email}>"],
      ["Errors-To",        list.email],
      ["To",               to],
      ["Subject",          subject],
      ["From",             "\"#{list.name}\" <#{list.email}>"],
      ["Sender",           list.email],
      ["Date",             Time.now.strftime(DATE_FORMAT)],
      ["Message-Id",       "<#{UUID.generate}@#{ENV['SENDER_HOST']}>"],
    )
    new.replace_headers!(*replace_headers) if replace_headers

    new.subject = subject
    new.to   = to
    new.from = list.email
    new.body = "#{body}#{list.signature}"
    new
  end

  # Create a email from scratch.
  # Create a response email for an existing one.
  #
  # @param list [Mailinglist]
  # @param mail [Protocols::Mail]
  def to_response(list:, mail:, body:)
    new = clone
    new.replace_headers!(
      ["User-Agent",       USER_AGENT],
      # require people to do not respond privatly
      ["Reply-to",         "#{list.name} <#{list.email}>"],
      ["Errors-To",        list.email],
      ["To",               list.email],
      ["From",             "\"#{list.name}\" <#{list.email}>"],
      ["Sender",           list.email],
      ["Date",             Time.now.strftime(DATE_FORMAT)],
      ["Message-Id",       "<#{UUID.generate}@#{ENV['SENDER_HOST']}>"],
    )

    new.to   = mail.from
    new.from = list.email
    new.body = "#{body}#{list.signature}"
    new
  end

  # :nodoc:
  def self.parse_rfc822_headers(rfc822)
    headers = []
    rfc822.split("\r\n").each do |line|
      if line[0] == "\t"
        headers.last[1] << "\r\n" << line
      else
        k, v = line.split(":", 2)
        headers << [k, v.to_s.strip]
      end
    end
    headers
  end

  # Get a header by key. Case insensitive. Only return the first occurence.
  #
  # @return [String?]
  def header(k)
    tuple = headers.find { _1[0].downcase == k.downcase }
    tuple && tuple[1]
  end

  # Getter the all the parsed headers.
  #
  # @return [Array(Tuple(String, String))]
  def headers
    @headers ||= []
  end

  # Generate a list of headers
  def kept_headers
    headers.filter { HEADERS_KEEP.include?(_1[0]) }
  end

  # Add new headers. Do not overwrite existing ones.
  #
  # @params new_headers is a [ [key, value], ...] array
  def add_headers!(*new_headers)
    new_headers.each { headers << _1 }
  end

  # Remove existing headers.
  #
  # @param headers_keys is a [word, word, ...] array
  def delete_headers!(*headers_keys)
    headers_keys.each { delete_header! _1 }
  end

  # :nodoc:
  private def delete_header!(header_key)
    headers.delete_if { _1[0].downcase == header_key.downcase }
  end

  # Remove conflicting headers and replace them with a new set.
  # If the header do not exist yet it is simply added.
  #
  # @params update_headers is a [ [key, value], ...] array
  def replace_headers!(*update_headers)
    delete_headers!(*update_headers.map(&:first))
    add_headers!(*update_headers)
  end

  # TODO
  # List-Id: <mailto:>
  # List-Help: <mailto:>
  # List-Subscribe: <mailto:>
  # List-Unsubscribe: <mailto:>
  # List-Post: <mailto:>
  # List-Owner: <mailto:>
  # List-Archive: <https://...>
  # Archived-At: <https://...>
  # List-Post: NO (posting not allowed on this list)

  # :nodoc:
  private def smtp_headers
    headers.map { _1.join(": ") }.join("\r\n")
  end

  # Generate a Truple(String, String, String) that can be sent
  # directly to the SMTP server using Net::SMTP or Protocols::SMTP.
  #
  # Notes:
  # - https://www.ietf.org/rfc/rfc2369.txt
  # - https://www.rfc-editor.org/rfc/rfc5983
  def to_smtp
    [
      "#{smtp_headers}\r\n\r\n#{@body}",
      @from,
      @to,
    ]
  end
end
